# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import collections
import datetime
import enum
import io
import re
import struct
import typing

from spnego._text import to_text
from spnego._version import __version__ as pyspnego_version


class NegotiateFlags(enum.IntFlag):
    """NTLM Negotiation flags.

    Used during NTLM negotiation to negotiate the capabilities between the client and server.

    .. _NEGOTIATE:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/99d90ff4-957f-4c8a-80e4-5bfe5a9a9832
    """

    key_56 = 0x80000000
    key_exch = 0x40000000
    key_128 = 0x20000000
    r1 = 0x10000000
    r2 = 0x08000000
    r3 = 0x04000000
    version = 0x02000000
    r4 = 0x01000000
    target_info = 0x00800000
    non_nt_session_key = 0x00400000
    r5 = 0x00200000
    identity = 0x00100000
    extended_session_security = 0x00080000
    target_type_share = 0x00040000  # Not documented in MS-NLMP
    target_type_server = 0x00020000
    target_type_domain = 0x00010000
    always_sign = 0x00008000
    local_call = 0x00004000  # Not documented in MS-NLMP
    oem_workstation_supplied = 0x00002000
    oem_domain_name_supplied = 0x00001000
    anonymous = 0x00000800
    r8 = 0x00000400
    ntlm = 0x00000200
    r9 = 0x00000100
    lm_key = 0x00000080
    datagram = 0x00000040
    seal = 0x00000020
    sign = 0x00000010
    netware = 0x00000008  # Not documented in MS-NLMP
    request_target = 0x00000004
    oem = 0x00000002
    unicode = 0x00000001

    @classmethod
    def native_labels(cls) -> typing.Dict["NegotiateFlags", str]:
        return {
            NegotiateFlags.key_56: "NTLMSSP_NEGOTIATE_56",
            NegotiateFlags.key_exch: "NTLMSSP_NEGOTIATE_KEY_EXCH",
            NegotiateFlags.key_128: "NTLMSSP_NEGOTIATE_128",
            NegotiateFlags.r1: "NTLMSSP_RESERVED_R1",
            NegotiateFlags.r2: "NTLMSSP_RESERVED_R2",
            NegotiateFlags.r3: "NTLMSSP_RESERVED_R3",
            NegotiateFlags.version: "NTLMSSP_NEGOTIATE_VERSION",
            NegotiateFlags.r4: "NTLMSSP_RESERVED_R4",
            NegotiateFlags.target_info: "NTLMSSP_NEGOTIATE_TARGET_INFO",
            NegotiateFlags.non_nt_session_key: "NTLMSSP_REQUEST_NON_NT_SESSION_KEY",
            NegotiateFlags.r5: "NTLMSSP_RESERVED_R5",
            NegotiateFlags.identity: "NTLMSSP_NEGOTIATE_IDENTITY",
            NegotiateFlags.extended_session_security: "NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY",
            NegotiateFlags.target_type_share: "NTLMSSP_TARGET_TYPE_SHARE - R6",
            NegotiateFlags.target_type_server: "NTLMSSP_TARGET_TYPE_SERVER",
            NegotiateFlags.target_type_domain: "NTLMSSP_TARGET_TYPE_DOMAIN",
            NegotiateFlags.always_sign: "NTLMSSP_NEGOTIATE_ALWAYS_SIGN",
            NegotiateFlags.local_call: "NTLMSSP_NEGOTIATE_LOCAL_CALL - R7",
            NegotiateFlags.oem_workstation_supplied: "NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED",
            NegotiateFlags.oem_domain_name_supplied: "NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED",
            NegotiateFlags.anonymous: "NTLMSSP_ANOYNMOUS",
            NegotiateFlags.r8: "NTLMSSP_RESERVED_R8",
            NegotiateFlags.ntlm: "NTLMSSP_NEGOTIATE_NTLM",
            NegotiateFlags.r9: "NTLMSSP_RESERVED_R9",
            NegotiateFlags.lm_key: "NTLMSSP_NEGOTIATE_LM_KEY",
            NegotiateFlags.datagram: "NTLMSSP_NEGOTIATE_DATAGRAM",
            NegotiateFlags.seal: "NTLMSSP_NEGOTIATE_SEAL",
            NegotiateFlags.sign: "NTLMSSP_NEGOTIATE_SIGN",
            NegotiateFlags.netware: "NTLMSSP_NEGOTIATE_NETWARE - R10",
            NegotiateFlags.request_target: "NTLMSSP_REQUEST_TARGET",
            NegotiateFlags.oem: "NTLMSSP_NEGOTIATE_OEM",
            NegotiateFlags.unicode: "NTLMSSP_NEGOTIATE_UNICODE",
        }


class AvId(enum.IntFlag):
    """ID for an NTLM AV_PAIR.

    These are the IDs that can be set as the `AvId` on an `AV_PAIR`_.

    .. _AV_PAIR:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/83f5e789-660d-4781-8491-5f8c6641f75e
    """

    eol = 0x0000
    nb_computer_name = 0x0001
    nb_domain_name = 0x0002
    dns_computer_name = 0x0003
    dns_domain_name = 0x0004
    dns_tree_name = 0x0005
    flags = 0x0006
    timestamp = 0x0007
    single_host = 0x0008
    target_name = 0x0009
    channel_bindings = 0x000A

    @classmethod
    def native_labels(cls) -> typing.Dict["AvId", str]:
        return {
            AvId.eol: "MSV_AV_EOL",
            AvId.nb_computer_name: "MSV_AV_NB_COMPUTER_NAME",
            AvId.nb_domain_name: "MSV_AV_NB_DOMAIN_NAME",
            AvId.dns_computer_name: "MSV_AV_DNS_COMPUTER_NAME",
            AvId.dns_domain_name: "MSV_AV_DNS_DOMAIN_NAME",
            AvId.dns_tree_name: "MSV_AV_DNS_TREE_NAME",
            AvId.flags: "MSV_AV_FLAGS",
            AvId.timestamp: "MSV_AV_TIMESTAMP",
            AvId.single_host: "MSV_AV_SINGLE_HOST",
            AvId.target_name: "MSV_AV_TARGET_NAME",
            AvId.channel_bindings: "MSV_AV_CHANNEL_BINDINGS",
        }


class AvFlags(enum.IntFlag):
    """MsvAvFlags for an AV_PAIR.

    These are the flags that can be set on the MsvAvFlags entry of an NTLM `AV_PAIR`_.

    .. _AV_PAIR:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/83f5e789-660d-4781-8491-5f8c6641f75e
    """

    none = 0x00000000
    constrained = 0x00000001
    mic = 0x00000002
    untrusted_spn = 0x00000004

    @classmethod
    def native_labels(cls) -> typing.Dict["AvFlags", str]:
        return {
            AvFlags.constrained: "AUTHENTICATION_CONSTRAINED",
            AvFlags.mic: "MIC_PROVIDED",
            AvFlags.untrusted_spn: "UNTRUSTED_SPN_SOURCE",
        }


class MessageType(enum.IntEnum):
    negotiate = 1
    challenge = 2
    authenticate = 3

    @classmethod
    def native_labels(cls) -> typing.Dict["MessageType", str]:
        return {
            MessageType.negotiate: "NEGOTIATE_MESSAGE",
            MessageType.challenge: "CHALLENGE_MESSAGE",
            MessageType.authenticate: "AUTHENTICATE_MESSAGE",
        }


def _get_payload_offset(b_data: memoryview, field_offsets: typing.List[int]) -> int:
    payload_offset = None

    for field_offset in field_offsets:
        offset = struct.unpack("<I", b_data[field_offset + 4 : field_offset + 8].tobytes())[0]
        if not payload_offset or (offset and offset < payload_offset):
            payload_offset = offset

    return payload_offset or len(b_data)


def _pack_payload(
    data: typing.Any,
    b_payload: bytearray,
    payload_offset: int,
    pack_func: typing.Optional[typing.Callable[[typing.Any], bytes]] = None,
) -> typing.Tuple[bytes, int]:
    if data:
        b_data = pack_func(data) if pack_func else data
    else:
        b_data = b""

    b_payload.extend(b_data)
    length = len(b_data)

    b_field = (struct.pack("<H", length) * 2) + struct.pack("<I", payload_offset)
    payload_offset += length

    return b_field, payload_offset


def _unpack_payload(
    b_data: memoryview,
    field_offset: int,
    unpack_func: typing.Optional[typing.Callable[[bytes], typing.Any]] = None,
) -> typing.Any:
    field_len = struct.unpack("<H", b_data[field_offset : field_offset + 2].tobytes())[0]
    if field_len:
        field_offset = struct.unpack("<I", b_data[field_offset + 4 : field_offset + 8].tobytes())[0]
        b_value = b_data[field_offset : field_offset + field_len].tobytes()

        return unpack_func(b_value) if unpack_func else b_value


class _NTLMMessageMeta(type):
    __registry: typing.Dict[int, typing.Type] = {}

    def __init__(
        cls,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        cls.__registry[getattr(cls, "MESSAGE_TYPE", 0)] = cls

    def __call__(
        cls,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> "_NTLMMessageMeta":
        if "_b_data" in kwargs:
            message_type = struct.unpack("<I", kwargs["_b_data"][8:12])[0]
            new_cls = cls.__registry[message_type]

        else:
            new_cls = cls

        return super(_NTLMMessageMeta, new_cls).__call__(*args, **kwargs)


class NTLMMessage(metaclass=_NTLMMessageMeta):
    """Base NTLM message class that defines the pack and unpack functions."""

    MESSAGE_TYPE = 0
    MINIMUM_LENGTH = 0

    def __init__(self, encoding: typing.Optional[str] = None, _b_data: typing.Optional[bytes] = None) -> None:
        self.signature = b"NTLMSSP\x00" + struct.pack("<I", self.MESSAGE_TYPE)
        self._encoding = encoding or "windows-1252"

        if _b_data:
            if len(_b_data) < self.MINIMUM_LENGTH:
                raise ValueError("Invalid NTLM %s raw byte length" % self.__class__.__name__)

            self._data = memoryview(bytearray(_b_data))

        else:
            self._data = memoryview(b"")

    def pack(self) -> bytes:
        """Packs the structure to bytes."""
        return self._data.tobytes()

    @staticmethod
    def unpack(b_data: bytes, encoding: typing.Optional[str] = None) -> "NTLMMessage":
        """Unpacks the structure from bytes."""
        return NTLMMessage(encoding=encoding, _b_data=b_data)


class Negotiate(NTLMMessage):
    """NTLM Negotiate Message

    This structure represents an NTLM `NEGOTIATE_MESSAGE`_ that can be serialized and deserialized to and from
    bytes.

    Args:
        flags: The `NegotiateFlags` that the client has negotiated.
        domain_name: The `DomainName` of the client authentication domain.
        workstation: The `Workstation` of the client.
        version: The `Version` of the client.
        encoding: The OEM encoding to use for text fields.
        _b_data: The raw bytes of the message to unpack from.

    .. _NEGOTIATE_MESSAGE:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/b34032e5-3aae-4bc6-84c3-c6d80eadf7f2
    """

    MESSAGE_TYPE = MessageType.negotiate
    MINIMUM_LENGTH = 32

    def __init__(
        self,
        flags: int = 0,
        domain_name: typing.Optional[str] = None,
        workstation: typing.Optional[str] = None,
        version: typing.Optional["Version"] = None,
        encoding: typing.Optional[str] = None,
        _b_data: typing.Optional[bytes] = None,
    ) -> None:
        super(Negotiate, self).__init__(encoding=encoding, _b_data=_b_data)

        if not _b_data:
            b_payload = bytearray()

            payload_offset = 32

            b_version = b""
            if version:
                flags |= NegotiateFlags.version
                b_version = version.pack()
            payload_offset = _pack_payload(b_version, b_payload, payload_offset)[1]

            b_domain_name = b""
            if domain_name:
                flags |= NegotiateFlags.oem_domain_name_supplied
                b_domain_name = domain_name.encode(self._encoding)
            b_domain_name_fields, payload_offset = _pack_payload(b_domain_name, b_payload, payload_offset)

            b_workstation = b""
            if workstation:
                flags |= NegotiateFlags.oem_workstation_supplied
                b_workstation = workstation.encode(self._encoding)
            b_workstation_fields = _pack_payload(b_workstation, b_payload, payload_offset)[0]

            b_data = bytearray(self.signature)
            b_data.extend(struct.pack("<I", flags))
            b_data.extend(b_domain_name_fields)
            b_data.extend(b_workstation_fields)
            b_data.extend(b_payload)

            self._data = memoryview(b_data)

    @property
    def flags(self) -> int:
        """The negotiate flags for the Negotiate message."""
        return struct.unpack("<I", self._data[12:16].tobytes())[0]

    @flags.setter
    def flags(self, value: int) -> None:
        self._data[12:16] = struct.pack("<I", value)

    @property
    def domain_name(self) -> typing.Optional[str]:
        """The name of the client authentication domain."""
        return to_text(_unpack_payload(self._data, 16), encoding=self._encoding, errors="replace", nonstring="passthru")

    @property
    def workstation(self) -> typing.Optional[str]:
        """The name of the client machine."""
        return to_text(_unpack_payload(self._data, 24), encoding=self._encoding, errors="replace", nonstring="passthru")

    @property
    def version(self) -> typing.Optional["Version"]:
        """The client NTLM version."""
        payload_offset = self._payload_offset

        # If the payload offset is at 40 or more then the Version, or at least empty bytes, is in the payload.
        if payload_offset >= 40:
            return Version.unpack(self._data[32:40].tobytes())

        else:
            return None

    @property
    def _payload_offset(self) -> int:
        """Gets the offset of the first payload value."""
        return _get_payload_offset(self._data, [16, 24])

    @staticmethod
    def unpack(b_data: bytes, encoding: typing.Optional[str] = None) -> "Negotiate":
        msg = NTLMMessage.unpack(b_data, encoding=encoding)
        if not isinstance(msg, Negotiate):
            raise ValueError("Input message was not a NTLM Negotiate message")

        return msg


class Challenge(NTLMMessage):
    """NTLM Challenge Message

    This structure represents an NTLM `CHALLENGE_MESSAGE`_ that can be serialized and deserialized to and from
    bytes.

    Args:
        flags: The `NegotiateFlags` that the client has negotiated.
        server_challenge: The random 64-bit `ServerChallenge` nonce.
        target_name: The name of the acceptor server.
        target_info: The variable length `TargetInfo` information.
        version: The `Version` of the server.
        encoding: The OEM encoding to use for text fields if `NTLMSSP_NEGOTIATE_UNICODE` was not supported.
        _b_data: The raw NTLM Challenge bytes to unpack from.

    .. _CHALLENGE_MESSAGE:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/801a4681-8809-4be9-ab0d-61dcfe762786
    """

    MESSAGE_TYPE = MessageType.challenge
    MINIMUM_LENGTH = 48

    def __init__(
        self,
        flags: int = 0,
        server_challenge: typing.Optional[bytes] = None,
        target_name: typing.Optional[str] = None,
        target_info: typing.Optional["TargetInfo"] = None,
        version: typing.Optional["Version"] = None,
        encoding: typing.Optional[str] = None,
        _b_data: typing.Optional[bytes] = None,
    ):
        super(Challenge, self).__init__(encoding=encoding, _b_data=_b_data)

        if _b_data:
            self._encoding = "utf-16-le" if self.flags & NegotiateFlags.unicode else self._encoding

        else:
            self._encoding = "utf-16-le" if flags & NegotiateFlags.unicode else self._encoding

            b_payload = bytearray()
            payload_offset = 48

            b_version = b""
            if version:
                flags |= NegotiateFlags.version
                b_version = version.pack()
            payload_offset = _pack_payload(b_version, b_payload, payload_offset)[1]

            b_target_name = b""
            if target_name:
                flags |= NegotiateFlags.request_target
                b_target_name = target_name.encode(self._encoding)
            b_target_name_fields, payload_offset = _pack_payload(b_target_name, b_payload, payload_offset)

            b_target_info = b""
            if target_info:
                flags |= NegotiateFlags.target_info
                b_target_info = target_info.pack()
            b_target_info_fields = _pack_payload(b_target_info, b_payload, payload_offset)[0]

            b_data = bytearray(self.signature)
            b_data.extend(b_target_name_fields)
            b_data.extend(struct.pack("<I", flags))
            b_data.extend(b"\x00" * 8)  # ServerChallenge, set after self._data is initialised.
            b_data.extend(b"\x00" * 8)  # Reserved
            b_data.extend(b_target_info_fields)
            b_data.extend(b_payload)

            self._data = memoryview(b_data)

            if server_challenge:
                self.server_challenge = server_challenge

    @property
    def target_name(self) -> typing.Optional[str]:
        """The name of the server authentication realm."""
        return to_text(_unpack_payload(self._data, 12), encoding=self._encoding, nonstring="passthru")

    @property
    def flags(self) -> int:
        """The negotiate flags supported by the server."""
        return struct.unpack("<I", self._data[20:24].tobytes())[0]

    @flags.setter
    def flags(self, value: int) -> None:
        self._data[20:24] = struct.pack("<I", value)

    @property
    def server_challenge(self) -> bytes:
        """The server's 8 byte nonce challenge."""
        return self._data[24:32].tobytes()

    @server_challenge.setter
    def server_challenge(self, value: bytes) -> None:
        if not value or len(value) != 8:
            raise ValueError("NTLM Challenge ServerChallenge must be 8 bytes long")

        self._data[24:32] = value

    @property
    def target_info(self) -> typing.Optional["TargetInfo"]:
        """The AV_PAIR structures generated by the server."""
        return _unpack_payload(self._data, 40, lambda d: TargetInfo.unpack(d))

    @property
    def version(self) -> typing.Optional["Version"]:
        """The server NTLM version."""
        payload_offset = self._payload_offset

        # If the payload offset is at 56 or more then the Version, or at least empty bytes, is in the payload.
        if payload_offset >= 56:
            return Version.unpack(self._data[48:56].tobytes())

        else:
            return None

    @property
    def _payload_offset(self) -> int:
        """Gets the offset of the first payload value."""
        return _get_payload_offset(self._data, [12, 40])

    @staticmethod
    def unpack(b_data: bytes, encoding: typing.Optional[str] = None) -> "Challenge":
        msg = NTLMMessage.unpack(b_data, encoding=encoding)
        if not isinstance(msg, Challenge):
            raise ValueError("Input message was not a NTLM Challenge message")

        return msg


class Authenticate(NTLMMessage):
    """NTLM Authentication Message

    This structure represents an NTLM `AUTHENTICATION_MESSAGE`_ that can be serialized and deserialized to and from
    bytes.

    Args:
        flags: The `NegotiateFlags` that the client has negotiated.
        lm_challenge_response: The `LmChallengeResponse` for the client's secret.
        nt_challenge_response: The `NtChallengeResponse` for the client's secret.
        domain_name: The `DomainName` for the client.
        username: The `UserName` for the cleint.
        workstation: The `Workstation` for the client.
        encrypted_session_key: The `EncryptedRandomSessionKey` for the set up context.
        version: The `Version` of the client.
        mic: The `MIC` for the authentication exchange.
        encoding: The OEM encoding to use for text fields if `NTLMSSP_NEGOTIATE_UNICODE` was not supported.
        _b_data: The raw NTLM Authenticate bytes to unpack from.

    .. _AUTHENTICATION_MESSAGE:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/033d32cc-88f9-4483-9bf2-b273055038ce
    """

    MESSAGE_TYPE = MessageType.authenticate
    MINIMUM_LENGTH = 64

    def __init__(
        self,
        flags: int = 0,
        lm_challenge_response: typing.Optional[bytes] = None,
        nt_challenge_response: typing.Optional[bytes] = None,
        domain_name: typing.Optional[str] = None,
        username: typing.Optional[str] = None,
        workstation: typing.Optional[str] = None,
        encrypted_session_key: typing.Optional[bytes] = None,
        version: typing.Optional["Version"] = None,
        mic: typing.Optional[bytes] = None,
        encoding: typing.Optional[str] = None,
        _b_data: typing.Optional[bytes] = None,
    ) -> None:
        super(Authenticate, self).__init__(encoding=encoding, _b_data=_b_data)

        if _b_data:
            self._encoding = "utf-16-le" if self.flags & NegotiateFlags.unicode else self._encoding

        else:
            self._encoding = "utf-16-le" if flags & NegotiateFlags.unicode else self._encoding

            b_payload = bytearray()

            payload_offset = 64

            # While MS server accept a blank version field, other implementations aren't so kind. No need to be strict
            # about it and only add the version bytes if it's present.
            b_version = b""
            if version:
                flags |= NegotiateFlags.version
                b_version = version.pack()
            payload_offset = _pack_payload(b_version, b_payload, payload_offset)[1]

            # MIC
            payload_offset = _pack_payload(b"\x00" * 16, b_payload, payload_offset)[1]

            b_lm_response_fields, payload_offset = _pack_payload(lm_challenge_response, b_payload, payload_offset)
            b_nt_response_fields, payload_offset = _pack_payload(nt_challenge_response, b_payload, payload_offset)
            b_domain_fields, payload_offset = _pack_payload(
                domain_name, b_payload, payload_offset, lambda d: d.encode(self._encoding)
            )
            b_username_fields, payload_offset = _pack_payload(
                username, b_payload, payload_offset, lambda d: d.encode(self._encoding)
            )
            b_workstation_fields, payload_offset = _pack_payload(
                workstation, b_payload, payload_offset, lambda d: d.encode(self._encoding)
            )
            if encrypted_session_key:
                flags |= NegotiateFlags.key_exch
            b_session_key_fields = _pack_payload(encrypted_session_key, b_payload, payload_offset)[0]

            b_data = bytearray(self.signature)
            b_data.extend(b_lm_response_fields)
            b_data.extend(b_nt_response_fields)
            b_data.extend(b_domain_fields)
            b_data.extend(b_username_fields)
            b_data.extend(b_workstation_fields)
            b_data.extend(b_session_key_fields)
            b_data.extend(struct.pack("<I", flags))
            b_data.extend(b_payload)

            self._data = memoryview(b_data)

            if mic:
                self.mic = mic

    @property
    def lm_challenge_response(self) -> typing.Optional[bytes]:
        """The LmChallengeResponse or None if not set."""
        return _unpack_payload(self._data, 12)

    @property
    def nt_challenge_response(self) -> typing.Optional[bytes]:
        """The NtChallengeResponse or None if not set."""
        return _unpack_payload(self._data, 20)

    @property
    def domain_name(self) -> typing.Optional[str]:
        """The domain or computer name hosting the user account."""
        return to_text(_unpack_payload(self._data, 28), encoding=self._encoding, nonstring="passthru")

    @property
    def user_name(self) -> typing.Optional[str]:
        """The name of the user to be authenticated."""
        return to_text(_unpack_payload(self._data, 36), encoding=self._encoding, nonstring="passthru")

    @property
    def workstation(self) -> typing.Optional[str]:
        """The name of the computer to which the user is logged on."""
        return to_text(_unpack_payload(self._data, 44), encoding=self._encoding, nonstring="passthru")

    @property
    def encrypted_random_session_key(self) -> typing.Optional[bytes]:
        """The client's encrypted random session key."""
        return _unpack_payload(self._data, 52)

    @property
    def flags(self) -> int:
        """The negotiate flags supported by the client and server."""
        return struct.unpack("<I", self._data[60:64].tobytes())[0]

    @flags.setter
    def flags(self, value: int) -> None:
        self._data[60:64] = struct.pack("<I", value)

    @property
    def version(self) -> typing.Optional["Version"]:
        """The client NTLM version."""
        payload_offset = self._payload_offset

        # If the payload offset is at 64 (no MIC or Version) or 80 (only MIC) then no version is present.
        if payload_offset not in [64, 80] and payload_offset >= 72:
            return Version.unpack(self._data[64:72].tobytes())

        else:
            return None

    @property
    def mic(self) -> typing.Optional[bytes]:
        """The MIC for the Authenticate message."""
        mic_offset = self._get_mic_offset()
        if mic_offset:
            return self._data.tobytes()[mic_offset : mic_offset + 16]

        else:
            return None

    @mic.setter
    def mic(self, value: bytes) -> None:
        if len(value) != 16:
            raise ValueError("NTLM Authenticate MIC must be 16 bytes long")

        mic_offset = self._get_mic_offset()
        if mic_offset:
            self._data[mic_offset : mic_offset + 16] = value
        else:
            raise ValueError("Cannot set MIC on an Authenticate message with no MIC present")

    @property
    def _payload_offset(self) -> int:
        """Gets the offset of the first payload value."""
        return _get_payload_offset(self._data, [12, 20, 28, 36, 44, 52])

    def _get_mic_offset(self) -> int:
        """Gets the offset of the MIC structure if present."""
        payload_offset = self._payload_offset

        # If the payload offset is 88 or more then we must have the Version (8 bytes) and the MIC (16 bytes) plus
        # any random data after that.
        if payload_offset >= 88:
            return 72

        # If the payload offset is between 80 and 88, then we should have just the MIC and no Version.
        elif payload_offset >= 80:
            return 64

        # Not enough room for a MIC between the minimum size and the payload offset.
        else:
            return 0

    @staticmethod
    def unpack(b_data: bytes, encoding: typing.Optional[str] = None) -> "Authenticate":
        msg = NTLMMessage.unpack(b_data, encoding=encoding)
        if not isinstance(msg, Authenticate):
            raise ValueError("Input message was not a NTLM Authenticate message")

        return msg


class FileTime(datetime.datetime):
    """Windows FILETIME structure.

    FILETIME structure representing number of 100-nanosecond intervals that have elapsed since January 1, 1601 UTC.
    This subclasses the datetime object to provide a similar interface but with the `nanosecond` attribute.

    Attrs:
        nanosecond (int): The number of nanoseconds (< 1000) in the FileTime. Note this only has a precision of up to
            100 nanoseconds.

    .. _FILETIME:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/2c57429b-fdd4-488f-b5fc-9e4cf020fcdf
    """

    _EPOCH_FILETIME = 116444736000000000  # 1970-01-01 as FILETIME.

    def __new__(cls, *args: typing.Any, **kwargs: typing.Any) -> "FileTime":
        ns = 0
        if "nanosecond" in kwargs:
            ns = kwargs.pop("nanosecond")

        dt = super(FileTime, cls).__new__(cls, *args, **kwargs)
        dt.nanosecond = ns

        return dt

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        super().__init__()
        self.nanosecond = getattr(self, "nanosecond", None) or 0

    @classmethod
    def now(cls, tz: typing.Optional[datetime.tzinfo] = None) -> "FileTime":
        """Construct a FileTime from the current time and optional time zone info."""
        return FileTime.from_datetime(datetime.datetime.now(tz=tz))

    @classmethod
    def from_datetime(cls, dt: datetime.datetime, ns: int = 0) -> "FileTime":
        """Creates a FileTime object from a datetime object."""
        return FileTime(
            year=dt.year,
            month=dt.month,
            day=dt.day,
            hour=dt.hour,
            minute=dt.minute,
            second=dt.second,
            microsecond=dt.microsecond,
            tzinfo=dt.tzinfo,
            nanosecond=ns,
        )

    def __str__(self) -> str:
        """Displays the datetime in ISO 8601 including the 100th nanosecond internal like .NET does."""
        fraction_seconds = ""

        if self.microsecond or self.nanosecond:
            fraction_seconds = self.strftime(".%f")

            if self.nanosecond:
                fraction_seconds += str(self.nanosecond // 100)

        timezone = "Z"
        if self.tzinfo:
            utc_offset = self.strftime("%z")
            timezone = "%s:%s" % (utc_offset[:3], utc_offset[3:])

        # strftime doesn't support dates < 1900 on Python 2.7
        return "{0}-{1:02d}-{2:02d}T{3:02d}:{4:02d}:{5:02d}{6}{7}".format(
            self.year, self.month, self.day, self.hour, self.minute, self.second, fraction_seconds, timezone
        )

    def pack(self) -> bytes:
        """Packs the structure to bytes."""
        # Make sure we are dealing with a timezone aware datetime
        utc_tz = datetime.timezone.utc
        utc_dt = typing.cast(datetime.datetime, self.replace(tzinfo=self.tzinfo if self.tzinfo else utc_tz))

        # Get the time since UTC EPOCH in microseconds
        td = utc_dt.astimezone(utc_tz) - datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=utc_tz)
        epoch_time_ms = td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6

        # Add the EPOCH_FILETIME to the microseconds since EPOCH and finally the nanoseconds part.
        ns100 = FileTime._EPOCH_FILETIME + (epoch_time_ms * 10) + (self.nanosecond // 100)

        return struct.pack("<Q", ns100)

    @staticmethod
    def unpack(b_data: bytes) -> "FileTime":
        """Unpacks the structure from bytes."""
        filetime = struct.unpack("<Q", b_data)[0]  # 100 nanosecond intervals since 1601-01-01.

        # Create a datetime object based on the filetime microseconds
        epoch_time_ms = (filetime - FileTime._EPOCH_FILETIME) // 10
        dt = datetime.datetime(1970, 1, 1) + datetime.timedelta(microseconds=epoch_time_ms)

        # Create the FileTime object from the datetime object and add the nanoseconds.
        ns = int(filetime % 10) * 100

        return FileTime.from_datetime(dt, ns=ns)


class NTClientChallengeV2:
    """NTLMv2 Client Challenge

    The `NTLMv2_CLIENT_CHALLENGE`_ structure defines the client challenge in the AUTHENTICATE_MESSAGE. This structure
    is only used when NTLMv2 authentication is configured and is transported in the NT Challenge Response.

    Args:
        time_stamp: The timestamp, defaults to the current time.
        client_challenge: The 8 byte nonce generated by the client.
        av_pairs: The TargetInfo AV_PAIRS for the client challenge.
        _b_data: Raw byte string for the NTClientChallengeV2, when set the other args are ignored.

    .. _NTLMv2_CLIENT_CHALLENGE:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/aee311d6-21a7-4470-92a5-c4ecb022a87b
    """

    def __init__(
        self,
        time_stamp: typing.Optional["FileTime"] = None,
        client_challenge: typing.Optional[bytes] = None,
        av_pairs: typing.Optional["TargetInfo"] = None,
        _b_data: typing.Optional[bytes] = None,
    ) -> None:
        if _b_data:
            if len(_b_data) < 32:
                raise ValueError("Invalid NTClientChallengeV2 raw byte length")

            self._data = memoryview(bytearray(_b_data))

        else:
            time_stamp = time_stamp or FileTime.now()
            av_pairs = av_pairs or TargetInfo()

            b_data = bytearray(b"\x01\x01\x00\x00\x00\x00\x00\x00")
            b_data.extend(time_stamp.pack())
            b_data.extend(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")
            b_data.extend(av_pairs.pack())

            self._data = memoryview(b_data)

            client_challenge = client_challenge or b"\x00" * 8
            self.challenge_from_client = client_challenge

    @property
    def resp_type(self) -> int:
        """The current response type version, must be set to 1."""
        return struct.unpack("B", self._data[:1].tobytes())[0]

    @resp_type.setter
    def resp_type(self, value: int) -> None:
        self._data[:1] = struct.pack("B", value)

    @property
    def hi_resp_type(self) -> int:
        """The maximum response type supported, must be set to 1."""
        return struct.unpack("B", self._data[1:2].tobytes())[0]

    @hi_resp_type.setter
    def hi_resp_type(self, value: int) -> None:
        self._data[1:2] = struct.pack("B", value)

    @property
    def time_stamp(self) -> "FileTime":
        """The current system time."""
        return FileTime.unpack(self._data[8:16].tobytes())

    @time_stamp.setter
    def time_stamp(self, value: "FileTime") -> None:
        self._data[8:16] = value.pack()

    @property
    def challenge_from_client(self) -> bytes:
        """8 byte client challenge."""
        return self._data[16:24].tobytes()

    @challenge_from_client.setter
    def challenge_from_client(self, value: bytes) -> None:
        if len(value) != 8:
            raise ValueError("NTClientChallengeV2 ChallengeFromClient must be 8 bytes long")
        self._data[16:24] = value

    @property
    def av_pairs(self) -> "TargetInfo":
        """The target info AV_PAIR structures."""
        return TargetInfo.unpack(self._data[28:].tobytes())

    def pack(self) -> bytes:
        """Packs the NTClientChallengeV2 to bytes."""
        return self._data.tobytes()

    @staticmethod
    def unpack(b_data: bytes) -> "NTClientChallengeV2":
        """Unpacks the raw bytes to the NTClientChallengeV2 structure."""
        return NTClientChallengeV2(_b_data=b_data)


class TargetInfo(collections.OrderedDict):
    """A collection of AV_PAIR structures for the TargetInfo field.

    The `AV_PAIR`_ structure defines an attribute/value pair and sequences of these pairs are using in the
    :class:`Challenge` and :class:`Authenticate` messages. The value for each pair depends on the AvId specified.
    Each value can be get/set/del like a normal dictionary where the key is the AvId of the value.

    .. _AV_PAIR:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/83f5e789-660d-4781-8491-5f8c6641f75e
    """

    _FIELD_TYPES = {
        "text": (
            AvId.nb_computer_name,
            AvId.nb_domain_name,
            AvId.dns_computer_name,
            AvId.dns_domain_name,
            AvId.dns_tree_name,
            AvId.target_name,
        ),
        "int32": (AvId.flags,),
        "struct": (AvId.timestamp, AvId.single_host),
    }

    def __setitem__(self, key: AvId, value: typing.Any) -> None:
        if isinstance(value, bytes):
            if key == AvId.timestamp:
                value = FileTime.unpack(value)
            elif key == AvId.single_host:
                value = SingleHost.unpack(value)

        super(TargetInfo, self).__setitem__(key, value)

    def pack(self) -> bytes:
        """Packs the structure to bytes."""
        b_data = io.BytesIO()

        for av_id, value in self.items():
            # MsvAvEOL should only be set at the end, will just ignore these entries.
            if av_id == AvId.eol:
                continue

            if av_id in self._FIELD_TYPES["text"]:
                b_value = value.encode("utf-16-le")
            elif av_id in self._FIELD_TYPES["int32"]:
                b_value = struct.pack("<I", value)
            elif av_id in self._FIELD_TYPES["struct"]:
                b_value = value.pack()
            else:
                b_value = value

            b_data.write(struct.pack("<HH", av_id, len(b_value)) + b_value)

        b_data.write(b"\x00\x00\x00\x00")  # MsvAvEOL
        return b_data.getvalue()

    @staticmethod
    def unpack(b_data: bytes) -> "TargetInfo":
        """Unpacks the structure from bytes."""
        target_info = TargetInfo()
        b_io = io.BytesIO(b_data)

        b_av_id = b_io.read(2)

        while b_av_id:
            av_id = struct.unpack("<H", b_av_id)[0]
            length = struct.unpack("<H", b_io.read(2))[0]
            b_value = b_io.read(length)

            value: typing.Any
            if av_id in TargetInfo._FIELD_TYPES["text"]:
                # All AV_PAIRS are UNICODE encoded.
                value = b_value.decode("utf-16-le")

            elif av_id in TargetInfo._FIELD_TYPES["int32"]:
                value = AvFlags(struct.unpack("<I", b_value)[0])

            elif av_id == AvId.timestamp:
                value = FileTime.unpack(b_value)

            elif av_id == AvId.single_host:
                value = SingleHost.unpack(b_value)

            else:
                value = b_value

            target_info[AvId(av_id)] = value
            b_av_id = b_io.read(2)

        return target_info


class SingleHost:
    """Single_Host_Data structure for NTLM TargetInfo entry.

    `Single_Host_Data`_ structure allows a client to send machine-specific information within an authentication
    exchange to services on the same machine. If the server and client platforms are different or if they are on
    different hosts, then the information MUST be ignores.

    Args:
        size: A 32-bit unsigned int that defines size of the structure.
        z4: A 32-bit integer value, currently set to 0.
        custom_data: An 8-byte platform-specific blob containing info only relevant when the client and server are on
            the same host.
        machine_id: A 32-byte random number created at computer startup to identify the calling machine.
        _b_data: Create a SingleHost object from the raw data byte string.

    Attributes:
        size: See args.
        z4: See args.
        custom_data: See args.
        machine_id: See args.

    .. _Single_Host_Data:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/f221c061-cc40-4471-95da-d2ff71c85c5b
    """

    def __init__(
        self,
        size: int = 0,
        z4: int = 0,
        custom_data: typing.Optional[bytes] = None,
        machine_id: typing.Optional[bytes] = None,
        _b_data: typing.Optional[bytes] = None,
    ) -> None:
        if _b_data:
            if len(_b_data) < 8:
                raise ValueError("SingleHost bytes must have at least a length of 8")
            self._data = memoryview(_b_data)

        else:
            self._data = memoryview(bytearray(48))
            self.size = size
            self.z4 = z4
            self.custom_data = custom_data or b"\x00" * 8
            self.machine_id = machine_id or b"\x00" * 32

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (bytes, SingleHost)):
            return False

        if isinstance(other, SingleHost):
            other = other.pack()

        return self.pack() == other

    @property
    def size(self) -> int:
        return struct.unpack("<I", self._data[:4].tobytes())[0]

    @size.setter
    def size(self, value: int) -> None:
        self._data[:4] = struct.pack("<I", value)

    @property
    def z4(self) -> int:
        return struct.unpack("<I", self._data[4:8].tobytes())[0]

    @z4.setter
    def z4(self, value: int) -> None:
        self._data[4:8] = struct.pack("<I", value)

    @property
    def custom_data(self) -> bytes:
        return self._data[8:16].tobytes() if len(self._data) >= 16 else b""

    @custom_data.setter
    def custom_data(self, value: bytes) -> None:
        if len(value) != 8:
            raise ValueError("custom_data length must be 8 bytes long")
        if len(self._data) < 16:
            raise ValueError("cannot set custom_data on a SingleHost with less than 16 bytes of data")

        self._data[8:16] = value

    @property
    def machine_id(self) -> bytes:
        return self._data[16:48].tobytes() if len(self._data) >= 48 else b""

    @machine_id.setter
    def machine_id(self, value: bytes) -> None:
        if len(value) != 32:
            raise ValueError("machine_id length must be 32 bytes long")
        if len(self._data) < 48:
            raise ValueError("cannot set machine_id on a SingleHost with less than 48 bytes of data")

        self._data[16:48] = value

    def pack(self) -> bytes:
        """Packs the structure to bytes."""
        return self._data.tobytes()

    @staticmethod
    def unpack(b_data: bytes) -> "SingleHost":
        """Creates a SignleHost object from raw bytes."""
        return SingleHost(_b_data=b_data)


class Version:
    """A structure contains the OS information.

    The `VERSION`_ structure contains operating system version information that SHOULD be ignored. This structure is
    used for debugging purposes only and its value does not affect NTLM message processing. It is populated in the NTLM
    messages only if `NTLMSSP_NEGOTIATE_VERSION` (`NegotiateFlags.version`) is negotiated.

    Args:
        major: See args. The 8-bit unsigned int for the version major part.
        minor: The 8-bit unsigned int for the version minor part.
        build: The 16-bit unsigned int for the version build part.
        revision: An 8-bit unsigned integer for the current NTLMSSP revision. This field SHOULD be `0x0F`.
        _b_data: Create a Version object from the raw data byte string.

    Attrs:
        major: See args.
        minor: See args.
        build: See args.
        reserved: A reserved 3-byte field that isn't used in the NTLM spec.
        revision: See args.

    .. _VERSION:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/b1a6ceb2-f8ad-462b-b5af-f18527c48175
    """

    def __init__(
        self,
        major: int = 0,
        minor: int = 0,
        build: int = 0,
        revision: int = 0x0F,
        _b_data: typing.Optional[bytes] = None,
    ) -> None:
        if _b_data:
            if len(_b_data) != 8:
                raise ValueError("Version bytes must have a length of 8")

            self._data = memoryview(_b_data)

        else:
            self._data = memoryview(bytearray(8))
            self.major = major
            self.minor = minor
            self.build = build
            self.revision = revision

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (bytes, Version)):
            return False

        if isinstance(other, Version):
            other = other.pack()

        return self.pack() == other

    def __len__(self) -> int:
        return 8

    @property
    def major(self) -> int:
        return struct.unpack("B", self._data[0:1].tobytes())[0]

    @major.setter
    def major(self, value: int) -> None:
        self._data[0:1] = struct.pack("B", value)

    @property
    def minor(self) -> int:
        return struct.unpack("B", self._data[1:2].tobytes())[0]

    @minor.setter
    def minor(self, value: int) -> None:
        self._data[1:2] = struct.pack("B", value)

    @property
    def build(self) -> int:
        return struct.unpack("<H", self._data[2:4].tobytes())[0]

    @build.setter
    def build(self, value: int) -> None:
        self._data[2:4] = struct.pack("<H", value)

    @property
    def reserved(self) -> bytes:
        return self._data[4:7].tobytes()

    @property
    def revision(self) -> int:
        return struct.unpack("B", self._data[7:8].tobytes())[0]

    @revision.setter
    def revision(self, value: int) -> None:
        self._data[7:8] = struct.pack("B", value)

    def __repr__(self) -> str:
        return "<{0}.{1} {2}.{3}.{4}.{5}>".format(
            type(self).__module__, type(self).__name__, self.major, self.minor, self.build, self.revision
        )

    def __str__(self) -> str:
        return "%s.%s.%s.%s" % (self.major, self.minor, self.build, self.revision)

    def pack(self) -> bytes:
        """Packs the structure to bytes."""
        return self._data.tobytes()

    @staticmethod
    def get_current() -> "Version":
        """Generates an NTLM Version structure based on the pyspnego package version."""
        versions = []
        for v in pyspnego_version.split(".", 3):
            if not v:
                continue

            match = re.match(r"^(\d+)", v)
            if match:
                versions.append(int(match.group(1)))

        versions += [0] * (3 - len(versions))

        return Version(major=int(versions[0]), minor=int(versions[1]), build=int(versions[2]))

    @staticmethod
    def unpack(b_data: bytes) -> "Version":
        """Creates a Version object from raw bytes."""
        return Version(_b_data=b_data)



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import abc
import dataclasses
import enum
import typing
import warnings

from spnego._credential import Credential
from spnego._text import to_text
from spnego.channel_bindings import GssChannelBindings
from spnego.exceptions import FeatureMissingError, NegotiateOptions, SpnegoError
from spnego.iov import BufferType, IOVBuffer, IOVResBuffer

F = typing.TypeVar("F", bound=typing.Callable[..., typing.Any])
NativeIOV = typing.TypeVar("NativeIOV", bound=typing.Any)
IOV = typing.Union[
    IOVBuffer,
    IOVResBuffer,
    typing.Tuple[typing.Union[BufferType, int], typing.Union[bool, bytes, int]],
    typing.Union[BufferType, int],
    bytes,
]


def split_username(username: typing.Optional[str]) -> typing.Tuple[typing.Optional[str], typing.Optional[str]]:
    """Splits a username and returns the domain component.

    Will split a username in the Netlogon form `DOMAIN\\username` and return the domain and user part as separate
    strings. If the user does not contain the `DOMAIN\\` prefix or is in the `UPN` form then then user stays the same
    and the domain is an empty string.

    Args:
        username: The username to split

    Returns:
        Tuple[Optional[str], Optional[str]]: The domain and username.
    """
    if username is None:
        return None, None

    domain: typing.Optional[str]
    if "\\" in username:
        domain, username = username.split("\\", 1)
    else:
        domain = None

    return to_text(domain, nonstring="passthru"), to_text(username, nonstring="passthru")


def wrap_system_error(error_type: typing.Type, context: typing.Optional[str] = None) -> typing.Callable[[F], F]:
    """Wraps a function that makes a native GSSAPI/SSPI syscall and convert native exceptions to a SpnegoError.

    Wraps a function that can potentially raise a WindowsError or GSSError and converts it to the common SpnegoError
    that is exposed by this library. This is to ensure the context proxy functions raise a common set of errors rather
    than a specific error for the provider. The underlying error is preserved in the SpnegoError if the user wishes to
    inspect that.

    Args:
        error_type: The native error type that need to be wrapped.
        context: An optional context message to add to the error if raised.
    """

    def decorator(func: F) -> F:
        def wrapper(*args: typing.Any, **kwargs: typing.Any) -> F:
            try:
                return func(*args, **kwargs)

            except error_type as native_err:
                raise SpnegoError(base_error=native_err, context_msg=context) from native_err

        return typing.cast(F, wrapper)

    return decorator


class WrapResult(typing.NamedTuple):
    """Result of the `wrap()` function."""

    data: bytes  #: The bytes of the wrapped data.
    encrypted: bool  #: Whether the data was encrypted (True) or not (False).


class IOVWrapResult(typing.NamedTuple):
    """Result of the `wrap_iov()` function."""

    buffers: typing.Tuple[IOVResBuffer, ...]  #: The wrapped IOV buffers.
    encrypted: bool  #: Whether the buffer data was encrypted (True) or not (False).


class WinRMWrapResult(typing.NamedTuple):
    """Result of the `wrap_winrm()` function."""

    header: bytes  #: The header of the wrapped result.
    data: bytes  #: The wrapped data included any padding.
    padding_length: int  #: The length of the bytes added to the data for padding.


class UnwrapResult(typing.NamedTuple):
    """Result of the `unwrap()` function."""

    data: bytes  #: The bytes of the unwrapped data.
    encrypted: bool  #: Whether the input data was encrypted (True) or not (False)
    qop: int  #: The Quality of Protection used for the encrypted data.


class IOVUnwrapResult(typing.NamedTuple):
    """Result of the `unwrap_iov()` function."""

    buffers: typing.Tuple[IOVResBuffer, ...]  #: The unwrapped IOV buffers.
    encrypted: bool  #: Whether the input buffers were encrypted (True) or not (False)
    qop: int  #: The Quality of Protection used for the encrypted buffers.


@dataclasses.dataclass(frozen=True)
class SecPkgContextSizes:
    """Sizes of important structures used for messages.

    This dataclass exposes the sizes of important structures used in message
    support functions like wrap, wrap_iov, sign, etc. Use
    :meth:`ContextReq.query_message_sizes` to retrieve this value for an
    authenticated context.

    Currently only ``header`` is exposed but other sizes may be added in the
    future if needed.

    Attributes:
        header: The size of the header/signature of a wrapped token. This
            corresponds to cbSecurityTrailer in SecPkgContext_Sizes in SSPI and
            the size of the allocated GSS_IOV_BUFFER_TYPE_HEADER IOV buffer.
    """

    header: int


class ContextReq(enum.IntFlag):
    none = 0x00000000

    # GSSAPI|SSPI flags
    delegate = 0x00000001
    mutual_auth = 0x00000002
    replay_detect = 0x00000004
    sequence_detect = 0x00000008
    confidentiality = 0x00000010
    integrity = 0x00000020
    # anonymous = 0x00000040  # TODO: Add support for anonymous auth.
    dce_style = 0x00001000
    identify = 0x00002000
    # Requires newer python-gssapi version to support https://github.com/pythongssapi/python-gssapi/pull/218
    delegate_policy = 0x00080000

    # Special flag that disables integrity/confidentiality on Kerberos/Negotiate
    # This should not be set with integrity or confidentiality.
    no_integrity = 0x10000000

    # mutual_auth | replay_detect | sequence_detect | confidentiality | integrity
    default = 0x00000002 | 0x00000004 | 0x00000008 | 0x00000010 | 0x00000020


class GSSMech(str, enum.Enum):
    ntlm = "1.3.6.1.4.1.311.2.2.10"
    spnego = "1.3.6.1.5.5.2"

    # Kerberos has been put under several OIDs over time, we should only be using 'kerberos'.
    kerberos = "1.2.840.113554.1.2.2"  # The actual Kerberos OID, this should be the one used.
    _ms_kerberos = "1.2.840.48018.1.2.2"
    _kerberos_draft = "1.3.5.1.5.2"
    _iakerb = "1.3.6.1.5.2"

    # Not implemented.
    kerberos_u2u = "1.2.840.113554.1.2.2.3"
    negoex = "1.3.6.1.4.1.311.2.2.30"

    @classmethod
    def native_labels(cls) -> typing.Dict[str, str]:
        return {
            GSSMech.ntlm: "NTLM",
            GSSMech.ntlm.value: "NTLM",
            GSSMech.spnego: "SPNEGO",
            GSSMech.spnego.value: "SPNEGO",
            GSSMech.kerberos: "Kerberos",
            GSSMech.kerberos.value: "Kerberos",
            GSSMech._ms_kerberos: "MS Kerberos",
            GSSMech._ms_kerberos.value: "MS Kerberos",
            GSSMech._kerberos_draft: "Kerberos (draft)",
            GSSMech._kerberos_draft.value: "Kerberos (draft)",
            GSSMech._iakerb: "IAKerberos",
            GSSMech._iakerb.value: "IAKerberos",
            GSSMech.kerberos_u2u: "Kerberos User to User",
            GSSMech.kerberos_u2u.value: "Kerberos User to User",
            GSSMech.negoex: "NEGOEX",
            GSSMech.negoex.value: "NEGOEX",
        }

    @property
    def common_name(self) -> str:
        if self.is_kerberos_oid:
            return "kerberos"

        return self.name

    @property
    def is_kerberos_oid(self) -> bool:
        """Determines if the mech is a Kerberos mech.

        Kerberos has been known under serveral OIDs in the past. This tells the caller whether the OID is one of those
        "known" OIDs.

        Returns:
            bool: Whether the mech is a Kerberos mech (True) or not (False).
        """
        return self in [GSSMech.kerberos, GSSMech._ms_kerberos, GSSMech._kerberos_draft, GSSMech._iakerb]

    @staticmethod
    def from_oid(oid: str) -> "GSSMech":
        """Converts an OID string to a GSSMech value.

        Converts an OID string to a GSSMech value if it is known.

        Args:
            oid: The OID as a string to convert from.

        Raises:
            ValueError: if the OID is not a known GSSMech.
        """
        for mech in GSSMech:
            if mech.value == oid:
                return mech
        else:
            raise ValueError("'%s' is not a valid GSSMech OID" % oid)


class ContextProxy(metaclass=abc.ABCMeta):
    """Base class for a authentication context.

    A base class the defined a common entry point for the various authentication context's that are used in this
    library. For a new context to be added it must implement the abstract functions in this class and translate the
    calls to what is required internally.

    Args:
        credentials: A list of credentials to use for authentication.
        hostname: The principal part of the SPN. This is required for Kerberos auth to build the SPN.
        service: The service part of the SPN. This is required for Kerberos auth to build the SPN.
        channel_bindings: The optional :class:`spnego.channel_bindings.GssChannelBindings` for the context.
        context_req: The :class:`spnego.ContextReq` flags to use when setting up the context.
        usage: The usage of the context, `initiate` for a client and `accept` for a server.
        protocol: The protocol to authenticate with, can be `ntlm`, `kerberos`, or `negotiate`. Not all providers
            support all three protocols as that is handled by :class:`SPNEGOContext`.
        options: The :class:`spnego.NegotiateOptions` that define pyspnego specific options to control the negotiation.

    Attributes:
        usage (str): The usage of the context, `initiate` for a client and `accept` for a server.
        protocol (str): The protocol to set the context up with; `ntlm`, `kerberos`, or `negotiate`.
        spn (str): The service principal name of the service to connect to.
        channel_bindings (spnego.channel_bindings.GssChannelBindings): Optional channel bindings to provide with the
            context.
        options (NegotiateOptions): The user specified negotiation options.
        context_req (ContextReq): The context requirements flags as an int value specific to the context provider.
    """

    def __init__(
        self,
        credentials: typing.List[Credential],
        hostname: typing.Optional[str],
        service: typing.Optional[str],
        channel_bindings: typing.Optional[GssChannelBindings],
        context_req: ContextReq,
        usage: str,
        protocol: str,
        options: NegotiateOptions,
    ) -> None:
        self.usage = usage.lower()
        if self.usage not in ["initiate", "accept"]:
            raise ValueError("Invalid usage '%s', must be initiate or accept" % self.usage)

        self.protocol = protocol.lower()
        if self.protocol not in ["ntlm", "kerberos", "negotiate", "credssp"]:
            raise ValueError("Invalid protocol '%s', must be ntlm, kerberos, negotiate, or credssp" % self.protocol)

        if self.protocol not in self.available_protocols(options=options):
            raise ValueError("Protocol %s is not available" % self.protocol)

        self._hostname = hostname
        self._service = service
        self.spn = None
        if service or hostname:
            self.spn = to_text("%s/%s" % (service if service else "HOST", hostname or "unspecified"))

        self.channel_bindings = channel_bindings
        self.options = NegotiateOptions(options)

        self.context_req = context_req  # Generic context requirements.
        self._context_req = 0  # Provider specific context requirements.
        for generic, provider in self._context_attr_map:
            if context_req & generic:
                self._context_req |= provider

        self._context_attr = 0  # Provider specific context attributes, set by self.step().

        # Whether the context is wrapped inside another context - set by NegotiateProxy.
        self._is_wrapped = False

        if options & NegotiateOptions.wrapping_iov and not self.iov_available():
            raise FeatureMissingError(NegotiateOptions.wrapping_iov)

    @property
    def username(self) -> None:
        warnings.warn("username is deprecated", category=DeprecationWarning)
        return None

    @property
    def password(self) -> None:
        warnings.warn("password is deprecated", category=DeprecationWarning)
        return None

    @classmethod
    def available_protocols(cls, options: typing.Optional[NegotiateOptions] = None) -> typing.List[str]:
        """A list of protocols that the provider can offer.

        Returns a list of protocols the underlying provider can implement. Currently only kerberos, negotiate, or ntlm
        is understood. The protocols that are available for each proxy context depend on the OS platform and what
        libraries are installed. See each proxy's `available_protocols` function for more info.

        Args:
            options: The context requirements of :class:`NegotiationOptions` that state what the client requires.

        Returns:
            List[str]: The list of protocols that the context can use.
        """
        return ["kerberos", "negotiate", "ntlm"]  # pragma: no cover

    @classmethod
    def iov_available(cls) -> bool:
        """Whether the context supports IOV wrapping and unwrapping.

        Will return a bool that states whether the context supports IOV wrapping or unwrapping. The NTLM protocol on
        Linux does not support IOV and some Linux gssapi implementations do not expose the extension headers for this
        function. This gives the caller a sane way to determine whether it can use :meth:`wrap_iov` or
        :meth:`unwrap_iov`.

        Returns:
            bool: Whether the context provider supports IOV wrapping and unwrapping (True) or not (False).
        """
        return True  # pragma: no cover

    @property
    @abc.abstractmethod
    def client_principal(self) -> typing.Optional[str]:
        """The principal that was used authenticated by the acceptor.

        The name of the client principal that was used in the authentication context. This is `None` when
        `usage='initiate'` or the context has not been completed. The format of the principal name is dependent on the
        protocol and underlying library that was used to complete the authentication.

        Returns:
            Optional[str]: The client principal name.
        """
        pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def complete(self) -> bool:
        """Whether the context has completed the authentication process.

        Will return a bool that states whether the authentication process has completed successfully.

        Returns:
            bool: The authentication process is complete (True) or not (False).
        """
        pass  # pragma: no cover

    @property
    def context_attr(self) -> ContextReq:
        """The context attributes that were negotiated.

        This is the context attributes that were negotiated with the counterpart server. These attributes are only
        valid once the context is fully established.

        Returns:
            ContextReq: The flags that were negotiated.
        """
        attr = 0
        for generic, provider in self._context_attr_map:
            if self._context_attr & provider:
                attr |= generic

        return ContextReq(attr)

    @property
    @abc.abstractmethod
    def negotiated_protocol(self) -> typing.Optional[str]:
        """The name of the negotiated protocol.

        Once the authentication process has compeleted this will return the name of the negotiated context that was
        used. For pure NTLM and Kerberos this will always be `ntlm` or `kerberos` respectively but for SPNEGO this can
        be either of those two.

        Returns:
            Optional[str]: The protocol that was negotiated, can be `ntlm`, `kerberos`, or `negotiate. Will be `None`
                for the acceptor until it receives the first token from the initiator. Once the context is establish
                `negotiate` will change to either `ntlm` or `kerberos` to reflect the protocol that was used by SPNEGO.
        """
        pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def session_key(self) -> bytes:
        """The derived session key.

        Once the authentication process is complete, this will return the derived session key. It is recommended to not
        use this key for your own encryption processes and is only exposed because some libraries use this key in their
        protocols.

        Returns:
            bytes: The derived session key from the authenticated context.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def new_context(self) -> "ContextProxy":
        """Creates a new security context.

        Creates a new security context based on the current credential and
        options of the current context. This is useful when needing to set up a
        new security context without having to retrieve the credentials again.

        Returns:
            ContextProxy: The new security context.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def query_message_sizes(self) -> SecPkgContextSizes:
        """Gets the important structure sizes for message functions.

        Will get the important sizes for the various message functions used by
        the current authentication context. This must only be called once the
        context has been authenticated.

        Returns:
            SecPkgContextSizes: The sizes for the current context.

        Raises:
            NoContextError: The security context is not ready to be queried.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def step(
        self,
        in_token: typing.Optional[bytes] = None,
        *,
        channel_bindings: typing.Optional[GssChannelBindings] = None,
    ) -> typing.Optional[bytes]:
        """Performs a negotiation step.

        This method performs a negotiation step and processes/generates a token. This token should be then sent to the
        counterpart context to continue the authentication process.

        This should not be called once :meth:`complete` is True as the security context is complete.

        For the initiator this is equivalent to `gss_init_sec_context`_ for GSSAPI and `InitializeSecurityContext`_ for
        SSPI.

        For the acceptor this is equivalent to `gss_accept_sec_context`_ for GSSAPI and `AcceptSecurityContext`_ for
        SSPI.

        Args:
            in_token: The input token to process (or None to process no input token).
            channel_bindings: Optional channel bindings ot use in this step. Will take priority over channel bindings
                set in the context if both are specified.

        Returns:
            Optional[bytes]: The output token (or None if no output token is generated.

        .. _gss_init_sec_context:
            https://tools.ietf.org/html/rfc2744.html#section-5.19

        .. _InitializeSecurityContext:
            https://docs.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-initializesecuritycontextw

        .. _gss_accept_sec_context:
            https://tools.ietf.org/html/rfc2744.html#section-5.1

        .. _AcceptSecurityContext:
            https://docs.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-acceptsecuritycontext
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def wrap(self, data: bytes, encrypt: bool = True, qop: typing.Optional[int] = None) -> WrapResult:
        """Wrap a message, optionally with encryption.

        This wraps a message, signing it and optionally encrypting it. The :meth:`unwrap` will unwrap a message.

        This is the equivalent to `gss_wrap`_ for GSSAPI and `EncryptMessage`_ for SSPI.

        The SSPI function's `EncryptMessage`_ is called with the following buffers::

            SecBufferDesc(SECBUFFER_VERSION, [
                SecBuffer(SECBUFFER_TOKEN, sizes.cbSecurityTrailer, b""),
                SecBuffer(SECBUFFER_DATA, len(data), data),
                SecBuffer(SECBUFFER_PADDING, sizes.cbBlockSize, b""),
            ])

        Args:
            data: The data to wrap.
            encrypt: Whether to encrypt the data (True) or just wrap it with a MIC (False).
            qop: The desired Quality of Protection (or None to use the default).

        Returns:
            WrapResult: The wrapped result which contains the wrapped message and whether it was encrypted or not.

        .. _gss_wrap:
            https://tools.ietf.org/html/rfc2744.html#section-5.33

        .. _EncryptMessage:
            https://docs.microsoft.com/en-us/windows/win32/secauthn/encryptmessage--general
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def wrap_iov(
        self,
        iov: typing.Iterable[IOV],
        encrypt: bool = True,
        qop: typing.Optional[int] = None,
    ) -> IOVWrapResult:
        """Wrap/Encrypt an IOV buffer.

        This method wraps/encrypts an IOV buffer. The IOV buffers control how the data is to be processed. Because
        IOV wrapping is an extension to GSSAPI and not implemented for NTLM on Linux, this method may not always be
        available to the caller. Check the :meth:`iov_available` property.

        This is the equivalent to `gss_wrap_iov`_ for GSSAPI and `EncryptMessage`_ for SSPI.

        Args:
            iov: A list of :class:`spnego.iov.IOVBuffer` buffers to wrap.
            encrypt: Whether to encrypt the message (True) or just wrap it with a MIC (False).
            qop: The desired Quality of Protection (or None to use the default).

        Returns:
            IOVWrapResult: The wrapped result which contains the wrapped IOVBuffer bytes and whether it was encrypted
                or not.

        .. _gss_wrap_iov:
            http://k5wiki.kerberos.org/wiki/Projects/GSSAPI_DCE

        .. _EncryptMessage:
            https://docs.microsoft.com/en-us/windows/win32/secauthn/encryptmessage--general
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def wrap_winrm(self, data: bytes) -> WinRMWrapResult:
        """Wrap/Encrypt data for use with WinRM.

        This method wraps/encrypts bytes for use with WinRM message encryption.

        Args:
            data: The data to wrap.

        Returns:
            WinRMWrapResult: The wrapped result for use with WinRM message encryption.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def unwrap(self, data: bytes) -> UnwrapResult:
        """Unwrap a message.

        This unwraps a message created by :meth:`wrap`.

        This is the equivalent to `gss_unwrap`_ for GSSAPI and `DecryptMessage`_ for SSPI.

        The SSPI function's `DecryptMessage`_ is called with the following buffers::

            SecBufferDesc(SECBUFFER_VERSION, [
                SecBuffer(SECBUFFER_STREAM, len(data), data),
                SecBuffer(SECBUFFER_DATA, 0, b""),
            ])

        Args:
            data: The data to unwrap.

        Returns:
            UnwrapResult: The unwrapped message, whether it was encrypted, and the QoP used.

        .. _gss_unwrap:
            https://tools.ietf.org/html/rfc2744.html#section-5.31

        .. _DecryptMessage:
            https://docs.microsoft.com/en-us/windows/win32/secauthn/decryptmessage--general
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def unwrap_iov(
        self,
        iov: typing.Iterable[IOV],
    ) -> IOVUnwrapResult:
        """Unwrap/Decrypt an IOV buffer.

        This method unwraps/decrypts an IOV buffer. The IOV buffers control how the data is to be processed. Because
        IOV wrapping is an extension to GSSAPI and not implemented for NTLM on Linux, this method may not always be
        available to the caller. Check the :meth:`iov_available` property.

        This is the equivalent to `gss_unwrap_iov`_ for GSSAPI and `DecryptMessage`_ for SSPI.

        Args:
            iov: A list of :class:`spnego.iov.IOVBuffer` buffers to unwrap.

        Returns:
            IOVUnwrapResult: The unwrapped buffer bytes, whether it was encrypted, and the QoP used.

        .. _gss_unwrap_iov:
            http://k5wiki.kerberos.org/wiki/Projects/GSSAPI_DCE

        .. _DecryptMessage:
            https://docs.microsoft.com/en-us/windows/win32/secauthn/decryptmessage--general
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def unwrap_winrm(self, header: bytes, data: bytes) -> bytes:
        """Unwrap/Decrypt a WinRM message.

        This method unwraps/decrypts a WinRM message. It handles the complexities of unwrapping the data and dealing
        with the various system library calls.

        Args:
            header: The header portion of the WinRM wrapped result.
            data: The data portion of the WinRM wrapped result.

        Returns:
            bytes: The unwrapped message.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def sign(self, data: bytes, qop: typing.Optional[int] = None) -> bytes:
        """Generates a signature/MIC for a message.

        This method generates a MIC for the given data. This is unlike wrap which bundles the MIC and the message
        together. The :meth:`verify` method can be used to verify a MIC.

        This is the equivalent to `gss_get_mic`_ for GSSAPI and `MakeSignature`_ for SSPI.

        Args:
            data: The data to generate the MIC for.
            qop: The desired Quality of Protection (or None to use the default).

        Returns:
            bytes: The MIC for the data requested.

        .. _gss_get_mic:
            https://tools.ietf.org/html/rfc2744.html#section-5.15

        .. _MakeSignature:
            https://docs.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-makesignature
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def verify(self, data: bytes, mic: bytes) -> int:
        """Verify the signature/MIC for a message.

        Will verify that the given MIC matches the given data. If the MIC does not match the given data, an exception
        will be raised. The :meth:`sign` method can be used to sign data.

        This is the equivalent to `gss_verify_mic`_ for GSSAPI and `VerifySignature`_ for SSPI.

        Args:
            data: The data to verify against the MIC.
            mic: The MIC to verify against the data.

        Returns:
            int: The QoP (Quality of Protection) used.

        .. _gss_verify_mic:
            https://tools.ietf.org/html/rfc2744.html#section-5.32

        .. _VerifySignature:
            https://docs.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-verifysignature
        """
        pass  # pragma: no cover

    # Internal properties/functions not for public use.

    @property
    @abc.abstractmethod
    def _context_attr_map(self) -> typing.List[typing.Tuple[ContextReq, int]]:
        """Map the generic ContextReq into the provider specific flags.

        Will return a list of tuples that give the provider specific flag value for the generic ContextReq that is
        exposed to end users.

        Returns:
            List[Tuple[ContextReq, int], ...]: A list of tuples where tuple[0] is the ContextReq flag and tuple[1] is
                the relevant provider specific flag for our common one.
        """
        pass  # pragma: no cover

    def get_extra_info(
        self,
        name: str,
        default: typing.Any = None,
    ) -> typing.Any:
        """Return information about the security context.

        Returns extra information about the security context that is not defined
        as part of the standard :class:`ContextProxy` attributes or properties.
        By default there is no context specific information and it's up to the
        sub classes to implement their own.

        These names can be queried for a CredSSP context.

            client_credential:
                Used on an `acceptor` CredSSP context and contains the delegated
                credential sent by the client to the server. This is only
                available once the context is complete otherwise the default
                value is returned. The types returned can be
                :class:`TSPasswordCreds`, :class:`TSSmartCardCreds`, or
                :class:`TSRemoteGuardCreds`.

            sslcontext:
                The :class:`ssl.SSLContext` instance used for the CredSSP
                context.

            ssl_object:
                The :class:`ssl.SSLObject` instance used for the CredSSP
                context.

            auth_stage - added in 0.5.0:
                A string representing that sub authentication stage being
                performed in the CredSSP authentication stepping. The value
                here is meant to be a human friendly representation and not
                something to be relied upon.

            protocol_version - added in 0.5.0:
                The CredSSP protocol version that was negotiated between the
                initiator and acceptor. This is the minimum version number
                offered by both parties once the Negotiate authentication stage
                is complete.

        Args:
            name: The name/id of the information to retrieve.
            default: The default value to return if the information is not
                available on the current context proxy.

        Args:
            name: The name/id of the information to retrieve.
            default: The default value to return if the information is not
                available on the current context proxy.

        Returns:
            The information requested or the default value specified if the
            information isn't found.
        """
        return default

    @property
    def _requires_mech_list_mic(self) -> bool:
        """Determine if the SPNEGO mechListMIC is required for the sec context.

        When Microsoft hosts deal with NTLM through SPNEGO it always wants the mechListMIC to be present when the NTLM
        authentication message contains a MIC. This goes against RFC 4178 as a mechListMIC shouldn't be required if
        NTLM was the preferred mech from the initiator but we can't do anything about that now. Because we exclusively
        use SSPI on Windows hosts, which does all the work for us, this function only matter for Linux hosts when this
        library manually creates the SPNEGO token.

        The function performs 2 operations. When called before the NTLM authentication message has been created it
        tells the gss-ntlmssp mech that it's ok to generate the MIC. When the authentication message has been created
        it returns a bool stating whether the MIC was present in the auth message and subsequently whether we need to
        include the mechListMIC in the SPNEGO token.

        See `mech_required_mechlistMIC in MIT KRB5`_ for more information about how MIT KRB5 deals with this.

        Returns:
            bool: Whether the SPNEGO mechListMIC needs to be generated or not.

        .. _mech_requires_mechlistMIC:
            https://github.com/krb5/krb5/blob/b2fe66fed560ae28917a4acae6f6c0f020156353/src/lib/gssapi/spnego/spnego_mech.c#L493
        """
        return False  # pragma: no cover

    def _build_iov_list(
        self, iov: typing.Iterable[IOV], native_convert: typing.Callable[[IOVBuffer], NativeIOV]
    ) -> typing.List[NativeIOV]:
        """Creates a list of IOV buffers for the native provider needed."""
        provider_iov: typing.List[NativeIOV] = []

        for entry in iov:
            data: typing.Optional[typing.Union[bytes, int, bool]]
            if isinstance(entry, tuple):
                if len(entry) != 2:
                    raise ValueError("IOV entry tuple must contain 2 values, the type and data, see IOVBuffer.")

                if not isinstance(entry[0], int):
                    raise ValueError("IOV entry[0] must specify the BufferType as an int")
                buffer_type = entry[0]

                if entry[1] is not None and not isinstance(entry[1], (bytes, int, bool)):
                    raise ValueError(
                        "IOV entry[1] must specify the buffer bytes, length of the buffer, or whether "
                        "it is auto allocated."
                    )
                data = entry[1] if entry[1] is not None else b""

            elif isinstance(entry, int):
                buffer_type = entry
                data = None

            elif isinstance(entry, bytes):
                buffer_type = BufferType.data
                data = entry

            else:
                raise ValueError("IOV entry must be a IOVBuffer tuple, int, or bytes")

            iov_buffer = IOVBuffer(type=BufferType(buffer_type), data=data)
            provider_iov.append(native_convert(iov_buffer))

        return provider_iov

    def _reset_ntlm_crypto_state(self, outgoing: bool = True) -> None:
        """Reset the NTLM crypto handles after signing/verifying the SPNEGO mechListMIC.

        `MS-SPNG`_ documents that after signing or verifying the mechListMIC, the RC4 key state needs to be the same
        for the mechListMIC and for the first message signed/sealed by the application. Because we use SSPI on Windows
        hosts which does all the work for us this function only matters for Linux hosts.

        Args:
            outgoing: Whether to reset the outgoing or incoming RC4 key state.

        .. _MS-SPNG:
            https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-spng/b87587b3-9d72-4027-8131-b76b5368115f
        """
        pass  # pragma: no cover
